/*
* Copyright 2017 Gregor Smith
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.*;
import java.util.stream.*;
import java.awt.image.*;

import greenfoot.*;


/**
* <p>Retrieves individual sprites from a sprite sheet
* <code>GreenfootImage</code>. Any size of image and any number of sprites is
* accepted. Each sprite is assumed to be spaced by the same amount.</p>
* <p>Individual sprites are lazily loaded by their coordinates through
* {@link #getSprite(int, int) getSprite}.</p>
*/
public class SpriteSheet {
    private GreenfootImage image;
    private GreenfootImage[][] sprites;
    private int spriteWidthOnImage, spriteHeightOnImage;
    private int spriteWidth, spriteHeight;
    private int rows, columns;

    /**
    * @param image The spritesheet image.
    * @param rows Number of sprite rows in <code>image</code>.
    * @param columns Number of sprite columns in <code>image</code>.
    */
    public SpriteSheet(GreenfootImage image, int rows, int columns) {
        this.image = image;
        this.columns = columns;
        this.rows = rows;

        spriteWidthOnImage = image.getWidth() / columns;
        spriteHeightOnImage = image.getHeight() / rows;

        sprites = new GreenfootImage[rows][columns];
    }

    /**
    * @param filename Filename of the spritesheet image.
    * @param rows Number of sprite rows in <code>image</code>.
    * @param columns Number of sprite columns in <code>image</code>.
    */
    public SpriteSheet(String filename, int rows, int columns) {
        this(new GreenfootImage(filename), rows, columns);
    }

    /**
    * <p>Resizes all sprites to <code>width</code> by <code>height</code>.
    * Sprites that have already been loaded are resized immediately. Sprites
    * that have not yet been loaded are resized when first accessed through
    * {@link #getSprite(int, int) getSprite}.</p>
    *
    * @param width New width of all sprites.
    * @param height New height of all sprites.
    */
    public void resizeAll(int width, int height) {
        spriteWidth = width;
        spriteHeight = height;
        for (GreenfootImage[] row: sprites) {
            for (GreenfootImage image: row) {
                if (image != null) {
                    image.scale(width, height);
                }
            }
        }
    }

    /**
    * @param column Column of the sprite to get.
    * @param row Row of the sprite to get.
    * @returns The sprite at <code>(column, row)</code>.
    */
    public GreenfootImage getSprite(int column, int row) {
        GreenfootImage sprite = sprites[row][column];
        if (sprite == null) {
            int x = column * spriteWidthOnImage;
            int y = row * spriteHeightOnImage;

            BufferedImage subImage = image.getAwtImage()
                .getSubimage(x, y, spriteWidthOnImage, spriteHeightOnImage);

            /*
            * A GreenfootImage cannot be created directly from a BufferedImage.
            * Instead a new GreenfootImage is created, and its underlying
            * BufferedImage retrieved. subImage is then drawn on this new
            * BufferedImage via a Graphics2D.
            */
            sprite = sprites[row][column] = new GreenfootImage(
                spriteWidthOnImage, spriteHeightOnImage
            );
            sprite.getAwtImage()
                .createGraphics()
                .drawImage(subImage, 0, 0, null);
            sprite.scale(getSpriteWidth(), getSpriteHeight());
        }
        return sprite;
    }

    /**
    * @param row Row of sprites to get.
    * @returns List of all sprites on row <code>row</code> of the image.
    */
    public List<GreenfootImage> getRow(int row) {
        /*
        * Just returning sprites[row] isn't good enough as sprites
        * on the row may not yet be initialised.
        */
        return IntStream.range(0, columns)
            .mapToObj(column -> getSprite(column, row))
            .collect(Collectors.toList());
    }

    /**
    * @param column Column of sprites to get.
    * @returns List of all sprites on column <code>column</code> of the image.
    */
    public List<GreenfootImage> getColumn(int column) {
        return IntStream.range(0, rows)
            .mapToObj(row -> getSprite(column, row))
            .collect(Collectors.toList());
    }

    /**
    * @returns The source spritesheet image.
    */
    public GreenfootImage getImage() {
        return image;
    }

    /**
    * @returns The width of each sprite tile on the source spritesheet image.
    *          This remains unchanged even if
    *          {@link #resizeAll(int, int) resizeAll} is called.
    */
    public int getSpriteWidthOnImage() {
        return spriteWidthOnImage;
    }

    /**
    * @returns The height of each sprite tile on the source spritesheet image.
    *          This remains unchanged even if
    *          {@link #resizeAll(int, int) resizeAll} is called.
    */
    public int getSpriteHeightOnImage() {
        return spriteHeightOnImage;
    }

    /**
    * @returns The width of a sprite as would be returned by
    *          {@link #getSprite(int, int) getSprite}. Can be set with
    *          {@link #resizeAll(int, int) resizeAll}.
    */
    public int getSpriteWidth() {
        return spriteWidth == 0 ? spriteWidthOnImage : spriteWidth;
    }

    /**
    * @returns The height of a sprite as would be returned by
    *          {@link #getSprite(int, int) getSprite}. Can be set with
    *          {@link #resizeAll(int, int) resizeAll}.
    */
    public int getSpriteHeight() {
        return spriteHeight == 0 ? spriteHeightOnImage : spriteHeight;
    }

    /**
    * @returns Number of sprite rows in the source image.
    */
    public int getRows() {
        return rows;
    }

    /**
    * @returns Number of sprite columns in the source image.
    */
    public int getColumns() {
        return columns;
    }
}
